import multiprocessing
import os
import shutil
import unittest

import numpy
import pandas as pd
import pkg_resources
import yaml

from watson.watson import Watson


class TestsWatson(unittest.TestCase):
    # def test_iatson(self):
    #     object_dir = TestsWatson.get_path("TIC25155310_[1,_2]")
    #     vetting_dir = TestsWatson.get_path("vetting_test")
    #     predictions, predictions_cal = Watson.run_iatson("TIC 25155310", 3.2899, 199, 1327.51, 6.082,
    #                       vetting_dir, object_dir + '/params_star.csv', object_dir + '/lc.csv', transits_mask=None, plot_inputs=False)
    #     self.assertAlmostEqual(numpy.nanmean(predictions), 0.1954, 3)
    #     self.assertAlmostEqual(numpy.nanstd(predictions), 0.18934455039188502, 3)
    #     self.assertAlmostEqual(numpy.nanmean(predictions_cal), 0.0925005359807983, 3)
    #     self.assertAlmostEqual(numpy.nanstd(predictions_cal), 0.10950608490582747, 3)

    def test_vetting_by_params(self):
        object_dir = TestsWatson.get_path("TIC25155310_[1,_2]")
        vetting_dir = object_dir + "/vetting_0/"
        try:
            Watson(object_dir, vetting_dir).vetting("TIC 25155310", 3.2899, 1327.51, 199,
                                                    6.082, [1, 2], 0.07571, cadence=120,
                                                    cpus=multiprocessing.cpu_count() // 2, clean=False)
            files_in_dir = os.listdir(vetting_dir)
            self.assertEqual(len(files_in_dir), 35)
        finally:
            if os.path.exists(vetting_dir):
                shutil.rmtree(vetting_dir, ignore_errors=False)

    def test_vetting_by_files(self):
        object_dir = TestsWatson.get_path("TIC25155310_[1,_2]")
        vetting_dir = object_dir + "/vetting_0/"
        try:
            Watson(object_dir, vetting_dir).vetting("TIC 25155310", 8.32, 1327.51, 199, 6.082, [1, 2], 0.07571,
                                       a_rstar=20, cadence=120, lc_file=object_dir + "/lc.csv",
                                       lc_data_file=object_dir + "/lc_data.csv",
                                       tpfs_dir=object_dir + "/tpfs",
                                       apertures_file=object_dir + "/apertures.yaml",
                                       star_file=object_dir + "params_star.csv",
                                       cpus=multiprocessing.cpu_count() // 2, clean=False, only_summary=True)
            files_in_dir = os.listdir(vetting_dir)
            self.assertEqual(len(files_in_dir), 9)
        finally:
            if os.path.exists(vetting_dir):
                shutil.rmtree(vetting_dir, ignore_errors=False)

    def test_fov_plots(self):
        object_dir = TestsWatson.get_path("TIC25155310_[1,_2]")
        vetting_dir = object_dir + "/vetting_0/"
        fov_dir = object_dir + "/fov"
        os.mkdir(fov_dir)
        try:
            with open(object_dir + "/apertures.yaml") as apertures_file:
                apertures = yaml.load(apertures_file, yaml.SafeLoader)
                apertures = apertures["sectors"]
                Watson(object_dir, vetting_dir).vetting_field_of_view(fov_dir, "TESS", "25155310", 120, 63.374706,
                                                                      -69.226593, [1, 2], "tpf", apertures,
                                                                      1)
                files_in_dir = os.listdir(fov_dir)
                self.assertEqual(len(files_in_dir), 4)
        finally:
            if os.path.exists(fov_dir):
                shutil.rmtree(fov_dir, ignore_errors=False)

    def test_vetting_by_files_with_fov(self):
        object_dir = TestsWatson.get_path("TIC25155310_[1,_2]")
        vetting_dir = object_dir + "/vetting_0/"
        try:
            Watson(object_dir, vetting_dir).vetting("TIC 25155310", 3.2899, 1327.51, 199, 6.082, [1, 2], 0.07571,
                                       a_rstar=20, cadence=120, lc_file=object_dir + "/lc.csv",
                                       lc_data_file=object_dir + "/lc_data.csv",
                                       tpfs_dir=object_dir + "/tpfs",
                                       apertures_file=object_dir + "/apertures.yaml",
                                       star_file=object_dir + "params_star.csv",
                                       cpus=multiprocessing.cpu_count() // 2, create_fov_plots=True,
                                       cadence_fov=120, ra=63.3739396231274, dec=-69.226822697583, clean=False)
            files_in_dir = os.listdir(vetting_dir)
            self.assertEqual(len(files_in_dir), 36)
        finally:
            if os.path.exists(vetting_dir):
                shutil.rmtree(vetting_dir, ignore_errors=False)

    def test_vetting_by_files_with_transits_list(self):
        object_dir = TestsWatson.get_path("TIC25155310_[1,_2]")
        vetting_dir = object_dir + "/vetting_0/"
        try:
            transits_list_df = pd.read_csv(object_dir + "/transits_stats.csv")
            transits_list_df = transits_list_df[transits_list_df["candidate"] == 0]
            Watson(object_dir, vetting_dir).vetting("TIC 25155310", 3.2899, 1327.51, 199, 6.082, [1, 2], 0.07571,
                                       a_rstar=20, cadence=120, lc_file=object_dir + "/lc.csv",
                                       lc_data_file=object_dir + "/lc_data.csv",
                                       tpfs_dir=object_dir + "/tpfs",
                                       apertures_file=object_dir + "/apertures.yaml",
                                       star_file=object_dir + "params_star.csv",
                                       cpus=multiprocessing.cpu_count() // 2,
                                       transits_list=transits_list_df.to_dict("list"), ra=63.3739396231274,
                                       dec=-69.226822697583, clean=False)
            files_in_dir = os.listdir(vetting_dir)
            self.assertEquals(len(files_in_dir), 33)
        finally:
            if os.path.exists(vetting_dir):
                shutil.rmtree(vetting_dir, ignore_errors=False)

    def test_create_report(self):
        object_dir = TestsWatson.get_path("TIC25155310_[1,_2]")
        vetting_dir = TestsWatson.get_path("vetting_report_test/")
        transits_list_df = pd.read_csv(object_dir + "/transits_stats.csv")
        transits_list_df = transits_list_df[transits_list_df["candidate"] == 0]
        transits_list_df = transits_list_df[transits_list_df["depth"].notnull()]
        transits_list_df = transits_list_df[transits_list_df["t0"].notnull()]
        transits_list = numpy.array(transits_list_df.to_dict("list")["t0"])
        transits_list = transits_list[~numpy.isnan(transits_list)]
        Watson(vetting_dir, vetting_dir).report("TIC 25155310", 63.3739396231274, -69.226822697583, 1327.51, 3.2899, 199, 6.082,
                                   transits_list, [2, 5, 7], None, None, None, None)
        files_in_dir = os.listdir(vetting_dir)
        self.assertEqual(len(files_in_dir), 20)

    @staticmethod
    def get_path(path):
        """
        Gets right path for tests environment
        :param path:
        :return: the real path of the test resource
        """
        return pkg_resources.resource_filename(__name__, path)


if __name__ == '__main__':
    unittest.main()



# import os
import datetime
import logging
import os
import pathlib
from astropy.coordinates import Angle
from reportlab.lib import colors
from reportlab.lib.pagesizes import A4
from reportlab.lib.units import cm
from reportlab.lib.enums import TA_CENTER, TA_JUSTIFY
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.platypus import BaseDocTemplate, PageTemplate, Frame, Paragraph, Spacer, PageBreak, \
    Image, Table, TableStyle, ListFlowable
from os import path
from astropy import units as u
import pandas as pd
import numpy as np

width, height = A4
resources_dir = path.join(path.dirname(__file__))


class Report:
    LOGO_IMAGE = resources_dir + "/resources/images/watson.png"

    def __init__(self, data_dir, file_name, object_id, ra, dec, t0, period, duration, depth, transit_t0s_list,
                 summary_list_t0s_indexes, v, j, h, k, with_tpfs=True, is_summary=False):
        self.data_dir = data_dir
        self.file_name = file_name
        self.object_id = object_id
        self.ra = ra
        self.dec = dec
        self.t0 = t0
        self.period = period
        self.duration = duration
        self.depth = depth
        self.transit_t0s_list = transit_t0s_list
        self.summary_list_t0s_indexes = summary_list_t0s_indexes
        self.v = v
        self.j = j
        self.h = h
        self.k = k
        self.with_tpfs = with_tpfs
        self.is_summary = is_summary

    @staticmethod
    def row_colors(df, table_object):
        data_len = len(df)
        for each in range(1, data_len + 1):
            if each % 2 == 1:
                bg_color = colors.whitesmoke
            else:
                bg_color = colors.lightgrey
            table_object.setStyle(TableStyle([('BACKGROUND', (0, each), (-1, each), bg_color)]))

    @staticmethod
    def metrics_row_colors(df, table_object):
        for index, row in df.iterrows():
            if not isinstance(row['passed'], str):
                if np.isnan(row['passed']):
                    bg_color = colors.yellow
                elif row['passed'] is False or row['passed'] == 0:
                    bg_color = colors.red
                elif row['passed'] is True or row['passed'] == 1:
                    bg_color = colors.lightgreen
                table_index = index + 1
                table_object.setStyle(TableStyle([('BACKGROUND', (0, table_index), (-1, table_index), bg_color),
                                              ('FONTSIZE', (0, table_index), (-1, table_index), 9),
                                              ('TOPPADDING', (0, table_index), (-1, table_index), 1),
                                              ('BOTTOMPADDING', (0, table_index), (-1, table_index), 1)]))

    def create_header(self, canvas, doc):
        canvas.saveState()

        # Logo:
        canvas.drawImage(self.LOGO_IMAGE, x=1.5 * cm, y=26.8 * cm, height=2 * cm, width=2 * cm,
                         preserveAspectRatio=True)

        # Title:
        object_id_text = 'WATSON Transits Validation Report: %s' % self.object_id
        canvas.setFont(psfontname="Helvetica", size=12)
        canvas.drawRightString(x=13.5 * cm, y=27.5 * cm, text=object_id_text)
        if doc.page == 1:
            object_id_text = '%s TRANSITS VALIDATION REPORT' % self.object_id
            canvas.setFont(psfontname="Helvetica-Bold", size=23)
            canvas.drawCentredString(x=10 * cm, y=25.5 * cm, text=object_id_text)

        # Report date:
        report_date = datetime.datetime.now().strftime("%a, %d %B %Y, %H:%M:%S")
        report_date_text = '%s' % report_date

        canvas.setFont("Helvetica", 9)
        canvas.drawRightString(20.5 * cm, 28 * cm, report_date_text)

        canvas.restoreState()

    def create_footer(self, canvas, doc):
        canvas.saveState()

        # if doc.page == 1:
        #     # Footer con superíndice:
        #     textobject = canvas.beginText()
        #     textobject.setTextOrigin(1.8 * cm, 2.1 * cm)
        #     textobject.setFont("Helvetica", 5)
        #     textobject.setRise(5)
        #     textobject.textOut('1 ')
        #     textobject.setRise(0)
        #     textobject.setFont("Helvetica", 7)
        #     pie_pagina = 'Three possible observability values are defined: 1 - Entire transit is required, ' \
        #                  '0.5 - Transit midtime and either ingress or egress at least are required,\n' \
        #                  '0.25 - Only ingress or egress are required, with moon constraints of % sº as minimum ' \
        #                  'distance for new moon and % sº as minimum distance for full moon\n' \
        #                  'and for the observatories listed in the Table 2.' % (self.min_dist, self.max_dist)
        #
        #     for line in pie_pagina.splitlines():
        #         textobject.textLine(line)
        #
        #     canvas.drawText(textobject)

        # Powered by:
        page = "Powered by ReportLab"
        canvas.setFont("Helvetica", 9)
        canvas.drawRightString(7 * cm, 0.5 * cm, page)

        # Page:
        page = "Page %s" % doc.page
        canvas.setFont("Helvetica", 9)
        canvas.drawRightString(20.5 * cm, 0.5 * cm, page)

        canvas.restoreState()

    def create_report(self):
        # Styles to be used
        styles = getSampleStyleSheet()
        styles.add(ParagraphStyle(name="ParagraphAlignCenter", alignment=TA_CENTER))
        styles.add(ParagraphStyle(name="ParagraphAlignJustify", alignment=TA_JUSTIFY))
        styles.wordWrap = 'LTR'
        table_style = TableStyle([('VALIGN', (0, 0), (-1, -1), 'MIDDLE'),
                                  ('ALIGN', (0, 0), (-1, -1), 'CENTER'),
                                  ('INNERGRID', (0, 0), (-1, -1), 0.25, colors.black),
                                  ('BOX', (0, 0), (-1, -1), 0.25, colors.black),
                                  ('FONTSIZE', (0, 0), (-1, -1), 10),
                                  ])
        # Content:
        story = [Spacer(1, 75)]
        introduction = '<font name="HELVETICA" size="9">This document is created by the WATSON report generator (' \
                       '<a href="https://github.com/PlanetHunters/watson" color="blue">https://github.com/PlanetHunters/watson</a>) ' \
                       'and focuses on the target star %s.</font>' % self.object_id
        story.append(Paragraph(introduction, styles["ParagraphAlignJustify"]))

        story.append(Spacer(1, 30))

        # Generamos la tabla 1 con los parámetros:
        tabla1_data = [['RA (deg)', 'Dec (deg)', 'V (mag)', 'J (mag)', 'H (mag)', 'K (mag)'],
                       [Angle(self.ra, u.deg).to_string(unit=u.hourangle, sep=':',
                                                        precision=2) if self.ra is not None else '-',
                        Angle(self.dec, u.deg).to_string(unit=u.deg, sep=':',
                                                         precision=2) if self.dec is not None else '-',
                        round(self.v, 2) if self.v is not None else '-',
                        round(self.j, 2) if self.j is not None else '-',
                        round(self.h, 2) if self.h is not None else '-',
                        round(self.k, 2) if self.k is not None else '-']]
        table1_colwidth = [3.5 * cm, 3.5 * cm, 2 * cm, 2 * cm, 2 * cm, 2 * cm]
        table1_number_rows = len(tabla1_data)
        tabla1 = Table(tabla1_data, table1_colwidth, table1_number_rows * [0.75 * cm])
        tabla1.setStyle(table_style)
        # Le damos el estilo alternando colores de filas:
        Report.row_colors(tabla1_data, tabla1)
        story.append(tabla1)
        table1_descripcion = '<font name="HELVETICA" size="9"><strong>Table 1: </strong>\
                        The proposed target parameters.</font>'
        story.append(Spacer(1, 5))
        story.append(Paragraph(table1_descripcion, styles["ParagraphAlignCenter"]))
        story.append(Spacer(1, 15))
        # Generamos la tabla 2 con los parámetros:
        tabla2_data = [['T0 (d)', 'Period (d)', 'Duration (h)', 'Depth (ppt)'],
                       [round(self.t0, 4),
                        round(self.period, 4),
                        round(self.duration / 60, 2),
                        round(self.depth, 3)]]
        table2_colwidth = [4 * cm, 4 * cm, 3.5 * cm, 3.5 * cm]
        table2_number_rows = len(tabla2_data)
        tabla2 = Table(tabla2_data, table2_colwidth, table2_number_rows * [0.75 * cm])
        tabla2.setStyle(table_style)
        Report.row_colors(tabla2_data, tabla2)
        story.append(tabla2)
        story.append(Spacer(1, 5))
        table2_descripcion = '<font name="HELVETICA" size="9"><strong>Table 2: </strong>' \
                             'The candidate parameters.</font>'
        story.append(Paragraph(table2_descripcion, styles["ParagraphAlignCenter"]))
        story.append(Spacer(1, 15))
        metrics_file = self.data_dir + "/metrics.csv"
        iatson_file = self.data_dir + "/iatson.csv"
        gpt_file = self.data_dir + '/gpt.csv'
        table_data = [['Metric', 'Value']]
        metrics_df = pd.DataFrame(columns=['metric', 'score', 'passed'])
        if os.path.exists(iatson_file):
            iatson_df = pd.read_csv(iatson_file)
            score_median = iatson_df['prediction'].mean(skipna=True)
            score_uncertainty = iatson_df['prediction'].std(skipna=True)
            if score_median > 0.99:
                passed = True
            elif score_median > 0.5:
                passed = np.nan
            else:
                passed = False
            metrics_df = pd.concat([metrics_df, pd.DataFrame.from_dict(
                {"metric": ["WATSON-NET"], 'score': [round(score_median, 4)], 'passed': [passed]}, orient='columns')], ignore_index=True)
            if score_uncertainty < 0.015:
                passed = True
            elif score_uncertainty < 0.1:
                passed = np.nan
            else:
                passed = False
            metrics_df = pd.concat([metrics_df, pd.DataFrame.from_dict(
                {"metric": ["WATSON-NET err"], 'score': [round(score_uncertainty, 4)], 'passed': [passed]}, orient='columns')], ignore_index=True)
        if os.path.exists(gpt_file):
            gpt_df = pd.read_csv(gpt_file)
            score_gpt = gpt_df.loc[0, 'prediction']
            content_gpt = gpt_df.loc[0, 'content']
            if score_gpt == 1:
                passed = True
            elif score_gpt == 0:
                passed = False
            else:
                passed = np.nan
            metrics_df = pd.concat([metrics_df, pd.DataFrame.from_dict(
                {"metric": ["GPT"], 'score': [score_gpt], 'passed': [passed]},
                orient='columns')], ignore_index=True)
        if os.path.exists(metrics_file):
            metrics_df = pd.concat([metrics_df, pd.read_csv(metrics_file)], ignore_index=True)
            metrics_df['passed'] = metrics_df['passed'].replace({'True': 1, 'False': 0})
            metrics_df['passed'] = metrics_df['passed'].replace({True: 1, False: 0})
            metrics_df['passed'] = pd.to_numeric(metrics_df['passed'], errors='coerce')
            for index, metric_row in metrics_df.iterrows():
                table_data.append([metric_row['metric'], round(metric_row['score'], 3)])
            table_colwidth = [4 * cm, 4 * cm, 3.5 * cm]
            table_number_rows = len(table_data)
            table = Table(table_data, table_colwidth, table_number_rows * [0.5 * cm])
            table.setStyle(table_style)
            Report.metrics_row_colors(metrics_df, table)
            story.append(table)
            story.append(Spacer(1, 5))
            table_descripcion = '<font name="HELVETICA" size="9"><strong>Table 3: </strong>' \
                                + 'The results of the numerical tests. Green cells mean acceptable values. Yellow means ' \
                                + 'inconclusive. Red represents problematic metrics.</font>'
            story.append(Paragraph(table_descripcion, styles["ParagraphAlignCenter"]))
            story.append(Spacer(1, 15))
        if os.path.exists(gpt_file):
            story.append(Paragraph('<font name="HELVETICA" size="9">GPT has been enabled to analyze this report. Its output is:</font>', styles["ParagraphAlignJustify"]))
            for gpt_explanation_paragraph in content_gpt.split('\n'):
                story.append(Paragraph('<font name="HELVETICA" size="9">' + gpt_explanation_paragraph + '</font>', styles["ParagraphAlignJustify"]))
            story.append(Spacer(1, 30))
        cadences_file = self.data_dir + "/folded_cadences.png"
        figure = 1
        if os.path.exists(cadences_file):
            story.append(Image(cadences_file, width=16 * cm, height=20 * cm))
            descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(figure) + ': </strong>' \
                                                                                            'Folded curve for all available cadences.</font>'
            story.append(Spacer(1, 5))
            story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
            story.append(Spacer(1, 15))
            figure = figure + 1
        # neighbours_file_index = 0
        # neighbours_file = self.data_dir + "/star_nb_" + (neighbours_file_index) + ".png"
        # figure = 1
        # while os.path.exists(neighbours_file):
        #     story.append(Image(neighbours_file, width=16 * cm, height=16 * cm))
        #     descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(figure) + ': </strong>' \
        #                          'Nearby stars folded plot with the same period than the candidate.</font>'
        #     story.append(Spacer(1, 5))
        #     story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
        #     story.append(Spacer(1, 15))
        #     neighbours_file_index = neighbours_file_index + 1
        #     neighbours_file = self.data_dir + "/star_nb_" + str(neighbours_file_index) + ".png"
        #     figure = figure + 1
        transit_depths_file = self.data_dir + "/transit_depths.png"
        if os.path.exists(transit_depths_file):
            story.append(Image(transit_depths_file, width=14 * cm, height=8 * cm))
            descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(figure) + ': </strong>' \
                                                                                            'The candidate single-transits depths plot.</font>'
            story.append(Spacer(1, 5))
            story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
            story.append(Spacer(1, 15))
            figure = figure + 1
        story.append(Image(self.data_dir + "/odd_even_folded_curves.png", width=10 * cm, height=14 * cm))
        descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(figure) + ': </strong>' \
                                                                                        'Above, the candidate folded at its found period for the found epoch. ' \
                                                                                        'Middle, the candidate secondary event plot. ' \
                                                                                        'Bottom, the candidate odd and even curves and their binned difference with its best fitted model.</font>'
        story.append(Spacer(1, 5))
        story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
        story.append(Spacer(1, 15))
        figure = figure + 1
        source_offsets_file = self.data_dir + '/source_offsets.png'
        if os.path.exists(source_offsets_file):
            story.append(Image(source_offsets_file, width=14 * cm, height=21 * cm))
            descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(figure) + ': </strong>' \
                                                                                            'Above, the computed target offset (red circle) from the differential image offset (cyan dot)' \
                                                                                            ' and the per-pixel BLS SNR offset (green dot). Middle left, the right ascension centroid shift with binning. ' \
                                                                                            'Middle right, the declination centroid shift with binning. ' \
                                                                                            'Bottom left, optical ghost diagnostic curve for core flux.' \
                                                                                            'Bottom right, optical ghost diagnostic curve for halo flux</font>'
            story.append(Spacer(1, 5))
            story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
            story.append(Spacer(1, 15))
            figure = figure + 1
        if not self.is_summary:
            for file in sorted(list(pathlib.Path(self.data_dir).glob('folded_tpf_*.png'))):
                story.append(Image(str(file), width=14 * cm, height=22 * cm))
                descripcion = '<font name="HELVETICA" size="9"><strong>Figure ' + str(figure) + '' \
                                                                                                ': </strong>Above, the TPF and per-pixel BLS SNR best fits. Bottom left, the per-pixel BLS SNR for each' \
                                                                                                ' pixel. Bottom right, the differential images SNR for each pixel. The target position is represented ' \
                                                                                                'by a red star and the TPF independent source offset is represented by a white plus.</font>'
                story.append(Spacer(1, 5))
                story.append(Paragraph(descripcion, styles["ParagraphAlignCenter"]))
                figure = figure + 1
            story.append(PageBreak())
        introduction = '<font name="HELVETICA" size="9">The next pages will contain each of the single-transits ' \
                       'vetting sheets with the next information: </font>'
        story.append(Paragraph(introduction, styles["ParagraphAlignJustify"]))
        story.append(ListFlowable(
            [Paragraph(s, styles["ParagraphAlignJustify"]) for s in [
                "<b>TOP-LEFT</b>: Plot with single transit photometry found in the analyzed curve (with momentum dumps, if any).",
                "<b>TOP-CENTER</b>: Plot with X-axis data drift (X-axis motion vs X-axis centroid offset) around the transit times.",
                "<b>TOP-RIGHT</b>: Plot with Y-axis data drift (Y-axis motion vs Y-axis centroid offset) around the transit times.",
                "<b>CENTER-LEFT</b>: Plot with SAP for used aperture vs SAP for smaller aperture around the transit times.",
                "<b>CENTER-CENTER</b>: Plot with smaller aperture over used aperture on the target.",
                "<b>CENTER-RIGHT</b>: Plot with single transit photometry found in the analyzed curve around the transit times.",
                "<b>BOTTOM</b>: Plot TPF flux measurements for each pixel around the transit times.",
            ]],
            bulletFormat="%s.",  # ←
        ))
        # Pasamos a la siguiente página:
        story.append(PageBreak())
        width = 17 * cm
        height = 24 * cm if self.with_tpfs else 10 * cm
        for index, transit_time in enumerate(self.transit_t0s_list):
            if self.summary_list_t0s_indexes is None or (self.summary_list_t0s_indexes is not None and
                                                         index in self.summary_list_t0s_indexes):
                image_file = self.data_dir + "/single_transit_" + str(index) + "_T0_" + str(transit_time) + ".png"
                if not os.path.exists(image_file):
                    logging.warning(f"The image {image_file} doesn't exist, skipping.")
                    continue
                image = Image(image_file, width=width, height=height)
                story.append(image)
                table3_descripcion = '<font name="HELVETICA" size="9"><strong>Figure %s: </strong>' \
                                     'The single transit no. %s vetting plots</font>' % (str(figure), str(index))
                story.append(Spacer(1, 5))
                story.append(Paragraph(table3_descripcion, styles["ParagraphAlignCenter"]))
                story.append(PageBreak() if self.with_tpfs or figure % 2 == 0 else Spacer(1, 5))
                figure = figure + 1

        # Construimos el documento:
        global_frame = Frame(1.5 * cm, 1.1 * cm, 18 * cm, 25.4 * cm, id='normal', showBoundary=0)
        global_template = PageTemplate(id='UnaColumna', frames=global_frame,
                                       onPage=self.create_header, onPageEnd=self.create_footer)
        doc = BaseDocTemplate(self.data_dir + "/" + self.object_id + "_" + self.file_name, pagesize=A4,
                              rightMargin=40, leftMargin=40,
                              topMargin=95, bottomMargin=15,
                              pageTemplates=global_template)
        doc.build(story)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

version = "0.12.1"
setuptools.setup(
    name="dearwatson", # Replace with your own username
    version=version,
    author="M. Dévora-Pajares",
    author_email="mdevorapajares@protonmail.com",
    description="Visual Vetting and Analysis of Transits from Space ObservatioNs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PlanetHunters/watson",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10',
    install_requires=[
                        "bokeh==2.4.2", # TPFPlotter dependency
                        'configparser==5.0.1',
                        "extension-helpers==0.1",
                        "exoml==0.1.26",
                        "imageio==2.9.0",
                        "openai==1.30.1",
                        'pyparsing==2.4.7', # Matplotlib dependency
                        "pyyaml==6.0.1",
                        "pillow==9.5.0",
                        "reportlab==4.0.7",
                        'setuptools>=41.0.0',
                        'triceratops==1.0.18'
    ]
)



import os
import sys
import os
import sys
import time

from lightkurve import search_targetpixelfile
from lightkurve import search_tesscut
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.colorbar import Colorbar
from matplotlib import patches
import numpy as np
import matplotlib.gridspec as gridspec
from bokeh.io import export_png
from bokeh.io.export import get_screenshot_as_png
import warnings
import numpy as np
from astropy.stats import sigma_clip
from astropy.coordinates import SkyCoord, Angle
import astropy.units as u
from astropy.visualization import SqrtStretch,LinearStretch
import astropy.visualization as stretching
from astropy.visualization.mpl_normalize import ImageNormalize
from astropy.table import Table, Column, MaskedColumn
from astropy.io import ascii
from astroquery.mast import Catalogs
import argparse

def cli():
    """command line inputs

    Get parameters from command line

    Returns
    -------
    Arguments passed by command line
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("tic", help="TIC number")
    parser.add_argument("-L", "--LIST", help="Only fit the LC", action="store_true")
    parser.add_argument("-S", "--SAVEGAIA", help="Save Gaia sources", action="store_true")
    parser.add_argument("-C", "--COORD", help="Use coordinates", default=False)
    parser.add_argument("--maglim", default=5., help="Maximum magnitude contrast respect to TIC")
    parser.add_argument("--sector", default=None, help="Select Sector if more than one")
    parser.add_argument("--gid", default=None, help="Gaia ID")
    parser.add_argument("--gmag", default=None, help="Gaia mag")
    args = parser.parse_args()
    return args

def add_gaia_figure_elements(tpf, magnitude_limit=18,targ_mag=10.):
    """Make the Gaia Figure Elements"""
    # Get the positions of the Gaia sources
    c1 = SkyCoord(tpf.ra, tpf.dec, frame='icrs', unit='deg')
    # Use pixel scale for query size
    pix_scale = 4.0  # arcseconds / pixel for Kepler, default
    if tpf.mission == 'TESS':
        pix_scale = 21.0
    # We are querying with a diameter as the radius, overfilling by 2x.
    from astroquery.vizier import Vizier
    Vizier.ROW_LIMIT = -1
    result = Vizier.query_region(c1, catalog=["I/345/gaia2"],
                                 radius=Angle(np.max(tpf.shape[1:]) * pix_scale, "arcsec"))
    no_targets_found_message = ValueError('Either no sources were found in the query region '
                                          'or Vizier is unavailable')
    too_few_found_message = ValueError('No sources found brighter than {:0.1f}'.format(magnitude_limit))
    if result is None:
        raise no_targets_found_message
    elif len(result) == 0:
        raise too_few_found_message
    result = result["I/345/gaia2"].to_pandas()
    result = result[result.Gmag < magnitude_limit]
    if len(result) == 0:
        raise no_targets_found_message
    radecs = np.vstack([result['RA_ICRS'], result['DE_ICRS']]).T
    coords = tpf.wcs.all_world2pix(radecs, 0.5) ## TODO, is origin supposed to be zero or one?
    year = ((tpf.astropy_time[0].jd - 2457206.375) * u.day).to(u.year)
    pmra = ((np.nan_to_num(np.asarray(result.pmRA)) * u.milliarcsecond/u.year) * year).to(u.arcsec).value
    pmdec = ((np.nan_to_num(np.asarray(result.pmDE)) * u.milliarcsecond/u.year) * year).to(u.arcsec).value
    result.RA_ICRS += pmra
    result.DE_ICRS += pmdec

    # Gently size the points by their Gaia magnitude
    sizes = 128.0 / 2**(result['Gmag']/targ_mag)#64.0 / 2**(result['Gmag']/5.0)
    one_over_parallax = 1.0 / (result['Plx']/1000.)
    r = (coords[:, 0]+tpf.column,coords[:, 1]+tpf.row,result['Gmag'])

    return r,result

# Plot orientation
def plot_orientation(tpf):
	"""
    Plot the orientation arrows

    Returns
    -------
    tpf read from lightkurve

	"""
	mean_tpf = np.mean(tpf.flux,axis=0)
	ny,nx = np.shape(mean_tpf)
	x0,y0 = tpf.column+int(0.9*nx),tpf.row+int(0.2*nx)
	# East
	tmp =  tpf.get_coordinates()
	ra00, dec00 = tmp[0][0][0][0], tmp[1][0][0][0]
	ra10,dec10 = tmp[0][0][0][-1], tmp[1][0][0][-1]
	theta = np.arctan((dec10-dec00)/(ra10-ra00))
	if (ra10-ra00) < 0.0: theta += np.pi
	#theta = -22.*np.pi/180.
	x1, y1 = 1.*np.cos(theta), 1.*np.sin(theta)
	plt.arrow(x0,y0,x1,y1,head_width=0.2,color='white')
	plt.text(x0+1.5*x1,y0+1.5*y1,'E',color='white')
	# North
	theta = theta +90.*np.pi/180.
	x1, y1 = 1.*np.cos(theta), 1.*np.sin(theta)
	plt.arrow(x0,y0,x1,y1,head_width=0.2,color='white')
	plt.text(x0+1.5*x1,y0+1.5*y1,'N',color='white')



def get_gaia_data(ra, dec):
    """
    Get Gaia parameters

    Returns
    -------
    RA, DEC
    """
    # Get the positions of the Gaia sources
    c1 = SkyCoord(ra, dec, frame='icrs', unit='deg')
    # We are querying with a diameter as the radius, overfilling by 2x.
    from astroquery.vizier import Vizier
    Vizier.ROW_LIMIT = -1
    result = Vizier.query_region(c1, catalog=["I/345/gaia2"],
                                 radius=Angle(10., "arcsec"))
    try:
    	result = result["I/345/gaia2"]
    except:
    	print 'Not in Gaia DR2. If you know the Gaia ID and Gmag, try the options --gid and --gmag.'
    	print 'Exiting without finishing...'
    	sys.exit()

    no_targets_found_message = ValueError('Either no sources were found in the query region '
                                          'or Vizier is unavailable')
    too_few_found_message = ValueError('No sources found closer than 1 arcsec to TPF coordinates')
    if result is None:
        raise no_targets_found_message
    elif len(result) == 0:
        raise too_few_found_message

    if len(result)>1:
        dist = np.sqrt((result['RA_ICRS']-ra)**2 + (result['DE_ICRS']-dec)**2)
        idx = np.where(dist == np.min(dist))[0][0]
        return result[idx]['Source'], result[idx]['Gmag']
    else:
        return result[0]['Source'], result[0]['Gmag']

def get_gaia_data_from_tic(tic):
    '''
    Get Gaia parameters

    Returns
    -----------------------
    GaiaID, Gaia_mag
    '''
    # Get the Gaia sources
    result = Catalogs.query_object('TIC'+tic, radius=.005, catalog="TIC")
    IDs = result['ID'].data.data
    k = np.where(IDs == tic)[0][0]
    GAIAs = result['GAIA'].data.data
    Gaiamags = result['GAIAmag'].data.data

    GAIA_k = GAIAs[k]
    Gaiamag_k = Gaiamags[k]

    if GAIA_k == '':
        GAIA_k = np.nan
    return GAIA_k, Gaiamag_k


def get_coord(tic):
	"""
	Get TIC corrdinates

	Returns
	-------
	TIC number
	"""
	try:
		catalog_data = Catalogs.query_object(objectname="TIC"+tic, catalog="TIC")
		ra = catalog_data[0]["ra"]
		dec = catalog_data[0]["dec"]
		return ra, dec
	except:
		print "ERROR: No gaia ID found for this TIC"


# ======================================
# 	        MAIN
# ======================================

if __name__ == "__main__":
	args = cli()
	if args.LIST:
		if args.COORD is not False:
			_tics = np.genfromtxt(args.tic,dtype=None)
			tics, ras, decs = [], [], []
			for t in _tics:
				tics.append(str(t[0]))
				ras.append(str(t[1]))
				decs.append(str(t[2]))
			ras, decs = np.array(ras), np.array(decs)
		else:
			_tics = np.genfromtxt(args.tic,dtype=None)
			tics = []
			for t in _tics: tics.append(str(t))
	else:
		if args.COORD:
			coords = args.COORD
			ras, decs = np.array([coords.split(',')[0]]), np.array([coords.split(',')[1]])
			tics = np.array([args.tic])
		else:
			tics = np.array([args.tic])


	for tt,tic in enumerate(tics):

		if args.COORD  is not False:
			ra,dec = ras[tt], decs[tt]
			print 'Working on '+tic+' (ra = '+ra+', '+'dec = '+dec+') ...'
		else:
			ra,dec = get_coord(tic)
			print 'Working on TIC'+tic+' (ra = '+str(ra)+', '+'dec = '+str(dec)+') ...'

		if args.gid != None:
			gaia_id, mag = args.gid, np.float(args.gmag)
		else:
			gaia_id, mag = get_gaia_data(ra, dec)
            gaia_id, mag = get_gaia_data_from_tic(tic)


		# By coordinates -----------------------------------------------------------------
		if args.COORD  is not False:
			                                                                             #
			if args.sector != None:
				tpf = search_tesscut(ra+" "+dec, sector=int(args.sector)).download(cutout_size=(12,12))     #

			else:
				tpf = search_tesscut(ra+" "+dec).download(cutout_size=(12,12))                             #
			pipeline = "False"
			print '    --> Using TESScut to get the TPF'

		# By TIC name --------------------------------------------------------------------
		else:
			# If the target is in the CTL (short-cadance targets)...
			try:
				if args.sector != None:
					tpf = search_targetpixelfile("TIC "+tic, sector=int(args.sector), mission='TESS').download()
					a = tpf.flux        # To check it has the flux array
					pipeline = "True"
				else:
					tpf = search_targetpixelfile("TIC "+tic, mission='TESS').download()
					a = tpf.flux        # To check it has the flux array
					pipeline = "True"

				print "    --> Target found in the CTL!"

			# ... otherwise if it still has a TIC number:
			except:
				if args.sector != None:
					tpf = search_tesscut("TIC "+tic, sector=int(args.sector)).download(cutout_size=(12,12))
				else:
					tpf = search_tesscut("TIC "+tic).download(cutout_size=(12,12))
				print "    -->  Target not in CTL. The FFI cut out was succesfully downloaded"
				pipeline = "False"

		fig = plt.figure(figsize=(6.93, 5.5))
		gs = gridspec.GridSpec(1,3, height_ratios=[1], width_ratios=[1,0.05,0.01])
		gs.update(left=0.05, right=0.95, bottom=0.12, top=0.95, wspace=0.01, hspace=0.03)
		ax1 = plt.subplot(gs[0,0])

		# TPF plot
		mean_tpf = np.mean(tpf.flux,axis=0)
		nx,ny = np.shape(mean_tpf)
		norm = ImageNormalize(stretch=stretching.LogStretch())
		division = np.int(np.log10(np.nanmax(tpf.flux)))
		splot = plt.imshow(np.nanmean(tpf.flux,axis=0)/10**division,norm=norm, \
						extent=[tpf.column,tpf.column+ny,tpf.row,tpf.row+nx],origin='bottom', zorder=0)

		# Pipeline aperture
		if pipeline == "True":                                           #
			aperture_mask = tpf.pipeline_mask
			aperture = tpf._parse_aperture_mask(aperture_mask)
			maskcolor = 'tomato'
			print "    --> Using pipeline aperture..."
		else:
			aperture_mask = tpf.create_threshold_mask(threshold=10,reference_pixel='center')
			aperture = tpf._parse_aperture_mask(aperture_mask)
			maskcolor = 'lightgray'
			print "    --> Using threshold aperture..."


		for i in range(aperture.shape[0]):
			for j in range(aperture.shape[1]):
				if aperture_mask[i, j]:
					ax1.add_patch(patches.Rectangle((j+tpf.column, i+tpf.row),
												   1, 1, color=maskcolor, fill=True,alpha=0.4))
					ax1.add_patch(patches.Rectangle((j+tpf.column, i+tpf.row),
												   1, 1, color=maskcolor, fill=False,alpha=1,lw=2))

		# Gaia sources
		r, res = add_gaia_figure_elements(tpf,magnitude_limit=mag+np.float(args.maglim),targ_mag=mag)
		x,y,gaiamags = r
		x, y, gaiamags=np.array(x)+0.5, np.array(y)+0.5, np.array(gaiamags)
		size = 128.0 / 2**((gaiamags-mag))
		plt.scatter(x,y,s=size,c='red',alpha=0.6, edgecolor=None,zorder = 10)

		# Gaia source for the target
		this = np.where(np.array(res['Source']) == int(gaia_id))[0]
		plt.scatter(x[this],y[this],marker='x',c='white',s=32,zorder = 11)

		# Legend
		fake_sizes = np.array([mag-2,mag,mag+2,mag+5, mag+8])
		for f in fake_sizes:
			size = 128.0 / 2**((f-mag))
			plt.scatter(0,0,s=size,c='red',alpha=0.6, edgecolor=None,zorder = 10,label = r'$\Delta m=$ '+str(round(f-mag,0)))

		ax1.legend(fancybox=True, framealpha=0.5)

		# Source labels
		dist = np.sqrt((x-x[this])**2+(y-y[this])**2)
		dsort = np.argsort(dist)
		for d,elem in enumerate(dsort):
			if dist[elem] < 6:
				plt.text(x[elem]-0.1,y[elem]+0.1,str(d+1),color='white', zorder=100)

		# Orientation arrows
		plot_orientation(tpf)

		# Labels and titles
		plt.xlim(tpf.column,tpf.column+ny)
		plt.ylim(tpf.row,tpf.row+nx)
		plt.xlabel('Pixel Column Number', fontsize=16)
		plt.ylabel('Pixel Row Number', fontsize=16)
		if args.COORD is not False:                                                                                          #
			plt.title('Coordinates '+tic+' - Sector '+str(tpf.sector), fontsize=16)# + ' - Camera '+str(tpf.camera))  #
		else:   												#
			plt.title('TIC '+tic+' - Sector '+str(tpf.sector), fontsize=16)# + ' - Camera '+str(tpf.camera))

		# Colorbar
		cbax = plt.subplot(gs[0,1]) # Place it where it should be.
		pos1 = cbax.get_position() # get the original position
		pos2 = [pos1.x0 - 0.05, pos1.y0 ,  pos1.width, pos1.height]
		cbax.set_position(pos2) # set a new position

		cb = Colorbar(ax = cbax, mappable = splot, orientation = 'vertical', ticklocation = 'right')
		plt.xticks(fontsize=14)
		exponent = r'$\times 10^'+str(division)+'$'
		cb.set_label(r'Flux '+exponent+r' (e$^-$)', labelpad=10, fontsize=16)

		plt.savefig('TPF_Gaia_TIC'+tic+'_S'+str(tpf.sector)+'.pdf')

		# Save Gaia sources info
		if args.SAVEGAIA:
			dist = np.sqrt((x-x[this])**2+(y-y[this])**2)
			GaiaID = np.array(res['Source'])
			srt = np.argsort(dist)
			x, y, gaiamags, dist, GaiaID = x[srt], y[srt], gaiamags[srt], dist[srt], GaiaID[srt]

			IDs = np.arange(len(x))+1
			inside = np.zeros(len(x))

			for i in range(aperture.shape[0]):
				for j in range(aperture.shape[1]):
					if aperture_mask[i, j]:
						xtpf, ytpf = j+tpf.column, i+tpf.row
						_inside = np.where((x > xtpf) & (x < xtpf+1) &
								 		   (y > ytpf) & (y < ytpf+1))[0]
						inside[_inside] = 1



			data = Table([IDs, GaiaID, x, y, dist, dist*21., gaiamags, inside.astype('int')],
						names=['# ID','GaiaID','x', 'y','Dist_pix','Dist_arcsec','Gmag', 'InAper'])
			ascii.write(data, 'Gaia_TIC'+tic+'_S'+str(tpf.sector)+'.dat',overwrite=True)
